/*
 * Copyright (c) 2016 Samsung Electronics Co., Ltd
 *
 * Licensed under the Flora License, Version 1.1 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://floralicense.org/license/
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <Elementary.h>
#include <efl_extension.h>
#include "$(appName).h"
#include "view.h"
#include "view_defines.h"

#define TRANSFORMATION_POINTS_CNT 4

#define DEG_90 90.0
#define DEG_360 360.0

static struct view_info {
	Evas_Object *win;
	Evas_Object *conform;
	Evas_Object *layout;
	Evas_Object *scroller;
	int current_page;
	int current_region;
} s_info = {
	.win = NULL,
	.conform = NULL,
	.layout = NULL,
	.scroller = NULL,
	.current_page = -1,
	.current_region = -1,
};

static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info);
static void _layout_back_cb(void *data, Evas_Object *obj, void *event_info);
static Eina_Bool _update_index_position_cb(void *data);
static void _page_change_start_cb(void *data, Evas_Object *obj, void *event_info);
static void _page_change_stop_cb(void *data, Evas_Object *obj, void *event_info);
static void _page_changed_cb(void *data, Evas_Object *obj, void *event_info);
static void _get_app_resource(const char *edj_file_in, char *edj_path_out);
static Evas_Object *_create_main_layout(Evas_Object *parent);
static Evas_Object *_create_scroller_layout(void);
static Evas_Object *_create_index_layout(void);
static Eina_Bool _add_scroller_page(Evas_Object *box, int r, int g, int b, const char *caption);
static int _get_page_count(void);
static int _get_page_current(void);
static Evas_Object *_get_indicator(Evas_Object *index, int c_page);
static void _rotate_indicator(Evas_Object *index, int indicator_no, double angle);

/**
 * @brief Creates essential objects: window, conformant and layout.
 * @param[in] user_data The data required for view module initialization.
 * @return This function returns 'EINA_TRUE' if the view is created successfully,
 * otherwise 'EINA_FALSE' is returned.
 */
Eina_Bool view_create(void *user_data)
{
	/* Create the window */
	s_info.win = view_create_win(PACKAGE);
	if (s_info.win == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to create a window.");
		return EINA_FALSE;
	}

	/* Create the conformant */
	s_info.conform = view_create_conformant(s_info.win);
	if (s_info.conform == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to create a conformant");
		return EINA_FALSE;
	}

	s_info.layout = _create_main_layout(s_info.conform);
	if (!s_info.layout)
		return EINA_FALSE;

	s_info.scroller = _create_scroller_layout();
	if (!s_info.scroller)
		return EINA_FALSE;

	if (!_create_index_layout())
		return EINA_FALSE;

	/* Show the window after main view is set up */
	evas_object_show(s_info.win);

	return EINA_TRUE;
}

/**
 * @brief Creates a basic window named package.
 * @param[in] pkg_name Name of the window.
 * @return This function returns the window object if it was created successfully,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_win(const char *pkg_name)
{
	Evas_Object *win = NULL;

	/*
	 * Window
	 * Create and initialize elm_win.
	 * elm_win is mandatory to manipulate the window.
	 */
	win = elm_win_util_standard_add(pkg_name, pkg_name);
	elm_win_conformant_set(win, EINA_TRUE);
	elm_win_autodel_set(win, EINA_TRUE);
	elm_win_indicator_mode_set(win, ELM_WIN_INDICATOR_HIDE);
	evas_object_smart_callback_add(win, "delete,request", _delete_win_request_cb, NULL);

	return win;
}

/**
 * @brief Creates a layout object for parent object based on provided EDJE script.
 * @param[in] parent The parent object for layout object.
 * @param[in] edj_file_name The relative path to the layout EDJE script.
 * @param[in] edj_group The name of the group to be loaded from the EDJE script.
 * @param[in] target_part_name The EDJE part's name where the layout is to be set.
 * @return The function returns layout object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_layout(Evas_Object *parent, const char *edj_file_name, const char *edj_group, const char *target_part_name)
{
	char edj_path[PATH_MAX] = {0, };
	Evas_Object *layout = NULL;

	if (!parent || !edj_file_name || !edj_group) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Wrong input arguments.");
		return NULL;
	}

	_get_app_resource(edj_file_name, edj_path);

	layout = elm_layout_add(parent);
	if (!layout) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_layout_add() failed.");
		return NULL;
	}

	if (!elm_layout_file_set(layout, edj_path, edj_group)) {
		dlog_print(DLOG_ERROR, LOG_TAG, "Function elm_layout_file_set() failed.");
		return NULL;
	}

	if (target_part_name)
		elm_object_part_content_set(parent, target_part_name, layout);
	else
		elm_object_content_set(parent, layout);

	evas_object_size_hint_weight_set(layout, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

	return layout;
}

/**
 * @brief Creates a conformant object.
 * @param[in] win The window object to which conformant is to be set.
 * Conformant is mandatory for base GUI to have proper size.
 * @return This function returns the conformant object if it was created successfully,
 * otherwise NULL is returned.
 */
Evas_Object *view_create_conformant(Evas_Object *win)
{
	/*
	 * Conformant
	 * Create and initialize elm_conformant.
	 * elm_conformant is mandatory for base GUI to have proper size
	 * when indicator or virtual keypad is visible.
	 */
	Evas_Object *conform = NULL;

	if (win == NULL) {
		dlog_print(DLOG_ERROR, LOG_TAG, "window is NULL.");
		return NULL;
	}

	conform = elm_conformant_add(win);
	evas_object_size_hint_weight_set(conform, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);
	elm_win_resize_object_add(win, conform);

	evas_object_show(conform);

	return conform;
}

/**
 * @brief Creates a scroller object for parent object.
 * @param[in] parent The parent object for scroller object.
 * @param[in] target_part_name The EDJE part's name where the scroller is to be set.
 * @return The function returns scroller object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_scroller(Evas_Object *parent, const char *target_part_name)
{
	Evas_Object *scroller = elm_scroller_add(parent);
	if (!scroller) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_scroller_add() failed");
		return NULL;
	}

	elm_scroller_page_relative_set(scroller, 1.0, 1.0);
	elm_scroller_content_min_limit(scroller, EINA_FALSE, EINA_FALSE);
	elm_scroller_bounce_set(scroller, EINA_FALSE, EINA_TRUE);
	elm_scroller_policy_set(scroller, ELM_SCROLLER_POLICY_OFF, ELM_SCROLLER_POLICY_OFF);
	elm_scroller_page_scroll_limit_set(scroller, 1, 1);
	elm_scroller_loop_set(scroller, EINA_TRUE, EINA_FALSE);

	if (target_part_name)
		elm_object_part_content_set(parent, target_part_name, scroller);

	evas_object_show(scroller);

	return scroller;
}

/**
 * @brief Creates a box object for parent object.
 * @param[in] parent The parent object for box object.
 * @param[in] target_part_name The EDJE part's name where the box is to be set.
 * @return The function returns box object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_box(Evas_Object *parent, const char *target_part_name)
{
	Evas_Object *box = elm_box_add(parent);
	if (!box) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_box_add() failed.");
		return NULL;
	}

	elm_box_horizontal_set(box, EINA_TRUE);
	elm_box_align_set(box, 0.5, 0.5);
	evas_object_size_hint_weight_set(box, EVAS_HINT_EXPAND, EVAS_HINT_EXPAND);

	if (target_part_name)
		elm_object_part_content_set(parent, target_part_name, box);

	evas_object_show(box);

	return box;
}

/**
 * @brief Creates a rectangle object for parent object.
 * @param[in] parent The parent object for rectangle object.
 * @param[in] target_part_name The EDJE part's name where the rectangle is to be set.
 * @return The function returns rectangle object if it was created successfully,
 * otherwise 'NULL' is returned.
 */
Evas_Object *view_create_rectangle(Evas_Object *parent, int width, int height, int r, int g, int b)
{
	Evas_Object *rect = evas_object_rectangle_add(evas_object_evas_get(parent));
	if (!rect) {
		dlog_print(DLOG_ERROR, LOG_TAG, "evas_object_rectangle_add() failed.");
		return NULL;
	}

	evas_object_color_set(rect, r, g, b, 255);

	evas_object_resize(rect, width, height);
	evas_object_size_hint_min_set(rect, width, height);

	return rect;
}

/**
 * @brief Destroys window and frees its resources.
 */
void view_destroy(void)
{
	if (s_info.win == NULL)
		return;

	evas_object_del(s_info.win);
}

/**
 * @brief Internal callback function invoked when the main window needs to be destroyed.
 * @param[in] data The user data passed to the evas_object_smart_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _delete_win_request_cb(void *data, Evas_Object *obj, void *event_info)
{
	ui_app_exit();
}

/**
 * @brief Internal callback function invoked on HW Back button press.
 * @param[in] data The user data passed to the eext_object_event_callback_add() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _layout_back_cb(void *data, Evas_Object *obj, void *event_info)
{
	/* Let window go to hide state. */
	elm_win_lower(s_info.win);
}

/**
 * @brief Internal callback function invoked when the indicator's bars needs to be rotated.
 * This callback function is attached with ecore_timer_add() function.
 * There are two bars rotated at once: bar indicating current page and adjacent bar indicating
 * next or previous page depending on page shift direction.
 * @param[in] data The user data passed to the callback attachment function.
 * @return This function returns EINA_FALSE in order to not invoke this callback again by ecore timer.
 */
static Eina_Bool _update_index_position_cb(void *data)
{
	int i = 0;
	int page_current = _get_page_current();
	int page_count = _get_page_count();
	Evas_Object *index = (Evas_Object *)data;

	if (!index) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_update_index_position_cb() failed. Invalid data argument.");
		return EINA_FALSE;
	}

	for (i = 0; i < page_count; i++)
		if (i == page_current)
			_rotate_indicator(index, i, 90.0);
		else
			_rotate_indicator(index, i, 0.0);

	return EINA_FALSE;
}

/**
 * @brief Internal callback function invoked when page dragging starts.
 * This callback function is attached with evas_object_smart_callback_add() function
 * for 'scroll,drag,start' event capturing.
 * @param[in] data The user data passed to the callback attachment() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _page_change_start_cb(void *data, Evas_Object *obj, void *event_info)
{
	s_info.current_page = _get_page_current();

	elm_scroller_region_get(obj, &s_info.current_region, NULL, NULL, NULL);
}

/**
 * @brief Internal callback function invoked when page dragging stops.
 * This callback function is attached with evas_object_smart_callback_add() function
 * for 'scroll,anim,stop' event capturing.
 * @param[in] data The user data passed to the callback attachment() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _page_change_stop_cb(void *data, Evas_Object *obj, void *event_info)
{
	s_info.current_page = -1;
	s_info.current_region = -1;

	_update_index_position_cb(data);
}

/**
 * @brief Internal callback function invoked when page is changed.
 * This callback function is attached with evas_object_smart_callback_add() function
 * for 'scroll' event capturing.
 * @param[in] data The user data passed to the callback attachment() function.
 * @param[in] obj The object invoking this callback function.
 * @param[in] event_info The structure containing the information on this event.
 */
static void _page_changed_cb(void *data, Evas_Object *obj, void *event_info)
{
	int x = 0;
	int w = 0;
	double angle = 0.0;
	int next_page = -1;
	int page_count = _get_page_count();
	Evas_Object *index = (Evas_Object *)data;

	elm_scroller_region_get(obj, &x, NULL, &w, NULL);

	angle = (double)(x - s_info.current_region) / w * DEG_90;
	if (abs(x - s_info.current_region) <= w) {
		next_page = x > s_info.current_region ? (s_info.current_page + 1) % page_count : s_info.current_page - 1;
	} else {
		if (x > s_info.current_region)
			next_page = s_info.current_page - 1;
		else
			next_page = 0;
	}

	if (next_page < 0)
		next_page = page_count - 1;

	_rotate_indicator(index, s_info.current_page, DEG_90 + angle);
	_rotate_indicator(index, next_page, DEG_360 + angle);
}

/**
 * @brief Internal function which creates fully qualified path to the provided resource file.
 * @param[in] edj_file_in The file name.
 * @param[out] edj_path_out The fully qualified path to the edj_file_in file.
 */
static void _get_app_resource(const char *edj_file_in, char *edj_path_out)
{
	char *res_path = app_get_resource_path();
	if (res_path) {
		snprintf(edj_path_out, PATH_MAX, "%s%s", res_path, edj_file_in);
		free(res_path);
	}
}

/**
 * @brief Internal function which creates the main layout.
 * @param[in] parent The parent object for the main layout object.
 * @return The function returns the main layout object if it was created successfully,
 * otherwise NULL is returned.
 */
static Evas_Object *_create_main_layout(Evas_Object *parent)
{
	Evas_Object *layout = view_create_layout(s_info.conform, EDJ_FILE, GRP_MAIN, "default");
	if (!layout)
		return NULL;

	eext_object_event_callback_add(layout, EEXT_CALLBACK_BACK, _layout_back_cb, NULL);
	evas_object_show(layout);

	return layout;
}

/**
 * @brief Internal function which creates the scroller layout including all pages.
 * @return The function returns the scroller layout if it was created successfully,
 * otherwise NULL is returned.
 */
static Evas_Object *_create_scroller_layout(void)
{
	Evas_Object *scroller;
	Evas_Object *box;

	scroller = view_create_scroller(s_info.layout, PART_SCROLLER);
	if (!scroller)
		return NULL;

	box = view_create_box(scroller, "default");
	if (!box)
		return NULL;

	_add_scroller_page(box, 255, 0, 0, "PAGE 1");
	_add_scroller_page(box, 0, 255, 0, "PAGE 2");
	_add_scroller_page(box, 0, 0, 255, "PAGE 3");
	_add_scroller_page(box, 0, 255, 255, "PAGE 4");

	return scroller;
}

/**
 * @brief Internal function which creates the page index layout and attaches
 * all the callback functions required for indices rotation on page shifting.
 * @return The function returns index layout if it was created successfully,
 * otherwise NULL is returned.
 */
static Evas_Object *_create_index_layout(void)
{
	Evas_Object *indicator;
	Evas_Object *index;
	int page_count = -1;
	int i = 0;

	page_count = _get_page_count();
	if (page_count < 0) {
		dlog_print(DLOG_ERROR, LOG_TAG, "failed to obtain pages count");
		return NULL;
	}

	index = view_create_box(s_info.layout, NULL);
	if (!index) {
		dlog_print(DLOG_ERROR, LOG_TAG, "box creation failed");
		return NULL;
	}

	for (i = 0 ; i < page_count; i++) {
		indicator = view_create_layout(index, EDJ_FILE, GRP_INDICATOR, NULL);
		elm_box_pack_end(index, indicator);
		evas_object_show(indicator);
	}

	/* This timer is created only for current page indicator state initialization. */
	ecore_timer_add(0.2, _update_index_position_cb, (void *)index);

	evas_object_smart_callback_add(s_info.scroller, "scroll", _page_changed_cb, (void *)index);
	evas_object_smart_callback_add(s_info.scroller, "scroll,drag,start", _page_change_start_cb, NULL);
	evas_object_smart_callback_add(s_info.scroller, "scroll,anim,stop", _page_change_stop_cb, (void *)index);

	elm_object_part_content_set(s_info.layout, PART_INDEX, index);

	return index;
}

/**
 * @brief Internal function which creates and adds a single page to the scroller layout.
 * @param[in] box The parent box to which the created page is inserted.
 * @param[in] r The red component of the color to be applied to the page background.
 * @param[in] g The green component of the color to be applied to the page background.
 * @param[in] b The blue component of the color to be applied to the page background.
 * @param[in] caption The text to be displayed on the newly created page.
 * @return The function returns 'EINA_TRUE' if the page was created successfully,
 * otherwise 'EINA_FALSE' is returned.
 */
static Eina_Bool _add_scroller_page(Evas_Object *box, int r, int g, int b, const char *caption)
{
	Evas_Object *page;
	Evas_Object *page_bg;
	int width = -1;
	int height = -1;

	page = view_create_layout(box, EDJ_FILE, GRP_PAGE, NULL);
	if (!page)
		return EINA_FALSE;

	elm_win_screen_size_get(s_info.win, NULL, NULL, &width, &height);
	page_bg = view_create_rectangle(box, width, (int)(height * SCROLLER_PAGE_HEIGHT), r, g, b);
	if (!page_bg)
		return EINA_FALSE;

	elm_object_part_content_set(page, PART_PAGE_BG, page_bg);
	elm_box_pack_end(box, page);

	if (caption)
		elm_object_part_text_set(page, PART_PAGE_CAPTION, caption);

	evas_object_show(page);

	return EINA_TRUE;
}

/**
 * @brief Internal function which gets the number of pages added to the scroller.
 * @return This function returns the number of pages contained in a scroller object.
 * If this function fail, the -1 value is returned.
 */
static int _get_page_count(void)
{
	Eina_List *list = NULL;
	Evas_Object *box = NULL;

	box = elm_object_content_get(s_info.scroller);
	if (!box) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_object_content_get() failed. Cannot get the pages quantity.");
		return -1;
	}

	list = elm_box_children_get(box);
	if (!list) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_box_children_get() failed. Cannot get the pages quantity.");
		return -1;
	}

	return eina_list_count(list);
}

/**
 * @brief Internal function which gets the number of currently displayed page.
 * @return This function returns the number of currently displayed page in a scroller object.
 * If this function fail, the -1 value is returned.
 */
static int _get_page_current(void)
{
	int page_num = -1;

	elm_scroller_current_page_get(s_info.scroller, &page_num, NULL);

	return page_num;
}

/**
 * @brief Internal function which gets an indicator object for given page number.
 * @param[in] index The page indicators container object.
 * @param[in] indicator_no The page number.
 * @return This function returns the indicator object for given page number
 * if it was successfully acquired, otherwise NULL is returned.
 */
static Evas_Object *_get_indicator(Evas_Object *index, int indicator_no)
{
	Eina_List *indices = elm_box_children_get(index);
	if (!indices) {
		dlog_print(DLOG_ERROR, LOG_TAG, "elm_box_children_get() failed. Cannot get indices list.");
		return NULL;
	}

	return (Evas_Object *)eina_list_nth(indices, indicator_no);
}

/**
 * @brief Internal function which rotates the indicator associated with given page
 * number. The rotation direction is defined by an angle value.
 * @param[in] index The page indicators container object.
 * @param[in] indicator_no The page number whose indicator is to be rotated.
 * @param[in] angle The rotation angle.
 */
static void _rotate_indicator(Evas_Object *index, int indicator_no, double angle)
{
	Evas_Object *indicator = NULL;
	Evas_Map *indicator_map = NULL;
	int x = -1;
	int y = -1;
	int w = -1;
	int h = -1;

	if (!index || indicator_no < 0 || angle < 0) {
		dlog_print(DLOG_ERROR, LOG_TAG, "_rotate_indicator() failed. Invalid arguments.");
		return;
	}

	indicator = _get_indicator(index, indicator_no);
	if (!indicator)
		return;

	evas_object_geometry_get(indicator, &x, &y, &w, &h);
	indicator_map = evas_map_new(TRANSFORMATION_POINTS_CNT);

	evas_map_util_points_populate_from_object(indicator_map, indicator);
	evas_map_util_rotate(indicator_map, angle, x + w / 2, y + w / 2);
	evas_object_map_set(indicator, indicator_map);
	evas_object_map_enable_set(indicator, EINA_TRUE);
}
